/* bootstrap.c
 *
 * Peforms bootstrap or rerandomization resampling analysis of
 * multiple treatments.
 *
 * written and copyright (c) by Garf and ff123, 2001
 * http://sjeng.sourceforge.net/
 * http://ff123.net/
 *
 *-------------------- GNU Lesser General Public License -------------------
 * http://www.gnu.org/copyleft/lesser.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Uses code written by Matthew Belmonte to perform student-t
 * calculations.  See copyright notice in fisher.c
 *
 * Uses code written by Makoto Matsumoto and Takuji Nishimura, and modified
 * by Shawn Cokus to perform random number generation.  See copyright
 * notice in random.c
 *
 *--------------------------------------------------------------------------
 *
 * 1a. Performs single-step, homscedastic bootstrap pairwise
 *     comparisons as described by Algorithm 3.1 of Westfall and
 *     Young, 1993.  The resampling algorithm is modified by
 *     default to treat each listener as a block, so that resampled
 *     listener values can only be drawn from the pool of values
 *     given by that listener.  Algorithm 3.1 draws from the entire
 *     pool of values to choose resampled values (the switch -unbl
 *     allows the program to draw resampled values from the entire
 *     pool).  Note that this program is restricted to balanced
 *     designs, where N for all groups are the same.  The adjustments
 *     necessary to perform bootstrap resmpling of block designs are
 *     described in section 4.2.2.
 *
 * 1b. Performs bootstrap free step-down p-value adjustment as
 *     described by Algorithm 3.3
 *
 * 2.  Performs single-step rerandomization as described by Algorithm
 *     3.4, except that each listener is treated as a block.  The
 *     current code cannot shuffle from the entire pool of codec x
 *     listener values.
 *
 * This program was written to evaluate the results of group
 * listening tests of various audio codecs and settings.
 *
 *---------------------- Version History ----------------------------
 *
 * v0.1    11-02-01
 *   Working version using rerandomization (permutation) resampling
 *     and free step-down p-value adjustment.  Strictly speaking,
 *     all p-value adjustments other than that of the minimum are not
 *     valid (pivotality condition is violated), so future algorithms
 *     will only allow single-step p-value adjustment when
 *     rerandomization resampling is performed (see p. 93 of Westfall
 *     and Young, 1993).
 *
 * v0.2    11-04-01
 *   Added command-line switches for specifying number of resamples
 *     and number of simulations (for adjusting p-values).
 *
 * v0.3    11-09-01
 *   Deprecating rerandomization resampling method in favor of
 *     bootstrap for now.  The -perm switch is used to perform
 *     the algorithm as implemented in version 0.1.  The
 *     rerandomization algorithm will be improved in the next version.
 *   Changed all floats to doubles, and changed DELTA to 1e-15.
 *     Slowdown is about 6% in the old rerandomization algorithm,
 *     but numbers in the calculation change a bit, so accuracy is
 *     improved.  Improved rerandomization algorithm in next version
 *     will drastically improve speed.
 *   Adding a new switch -n, for number of bootstraps (default = 10000).
 *   Program now performs single-step bootstrap adjustment of p-values.
 *
 * v0.4    11-11-01
 *   Removed old rerandomization code, and substituted new.
 *     Rerandomization using unblocked mode (values can be chosen
 *     from entire codec x listener pool) is not implemented.
 *   Now calculate unadjusted p-value matrix for blocked designs as
 *     default instead of the unblocked p-value matrix.
 *   Now calculate bootstrap free step-down, which is also the default
 *     p-value adjustment method.
 * v0.4a   11-11-01
 *   Forgot to enforce monoticity with free step-down.  Fixed.
 * v0.4b   11-11-01
 *   Aargh.  Broke single step-down when enforcing monoticity for free
 *     step-down.  Fixed.
 * v0.5    11-18-01
 *   Code released under LGPL, as mandated by the use of the Mersenne
 *     Twister library.
 *   Bug fix:  was losing power when using the default blocked mode
 *     because the resampled t-value calculations were using the
 *     means of the original data instead of the resampled data.
 *     Fixed.
 *
 * -------------------------------------------------------------------
 *
 * The input file must have the following file format:
 *
 * % Example input file
 * % Lines that start with a percent sign are designated as comments.
 * % Also, blank lines are ignored as comments.
 *
 * SampleA  SampleB  SampleC  SampleD
 *   2.4      4.0      3.2      3.2
 *   1.3      3.4      5.0      4.0
 *   4.0      4.0      2.0      2.0
 *
 * The first non-comment line contains the labels for the samples.
 * The labels must not contain any spaces in them.  The labels will be
 * truncated to MAXLABEL characters.
 *
 * -------------------------------------------------------------------
 *
 * References: "Resampling-Based Multiple Testing: Examples and Methods
 *               for p-value Adjustment," by Peter H. Westfall and S.
 *               Stanley Young, 1993, John Wiley & Sons, Inc., ISBN
 *               0-471-55761-7
 *             "Sensory Evaluation Techniques, 3rd Ed.," by Meilgaard,
 *               Civille, and Carr, 1999, CRC Press LLC, ISBN
 *               0-8493-0276-5
 *********************************************************************/
#include "bootstrap.h"


/**********************************************************************/
/* defines global variables
 */
void init(void)
{
  npadj = 10000;        /* number of resamples to derive adjusted p */
  adjustmode = FREE;    /* p-value adjustment mode */
  unblocked = FALSE;    /* type of resampling */
  resamplemode = BOOT;  /* resampling mode */
}

/**********************************************************************/
/* calculates the means of the treatments.
 */
void treatmeans(double data[MAXTREAT][MAXSAMPLE],
                double treatmeans[MAXTREAT])
{
  int i, j;   /* index variables */
  double sum;  /* sum of data */

  for (i=0; i<numtreat; i++)
    {
      sum = 0.0;
      for (j=0; j<numsample; j++)
        {
          sum += data[i][j];
        }
      treatmeans[i] = sum / numsample;
    }
  return;
}

/**********************************************************************/
/* calculates the residual mean of the data
 */
double residualmean(double data[MAXTREAT][MAXSAMPLE],
                    double means[MAXTREAT])
{
  int i, j;          /* index variables */
  double sum = 0.0;
  double residual;   /* result */

  for (i=0; i<numtreat; i++)
    {
      for (j=0; j<numsample; j++)
        {
          sum += pow((data[i][j] - means[i]), 2);
        }
    }
  residual = sqrt( sum/((numsample-1)*numtreat) );

  return residual;
}

/**********************************************************************/
/* calculates the mean square of the error
 */
double meansquareerror(double data[MAXTREAT][MAXSAMPLE])
{
  int i, j;                    /* index variables */
  double sum = 0;              /* sum of data */
  double mse;                  /* mean square of the error */
  double mean;                 /* mean of data */
  double ss_tot = 0;           /* devsq of the data */
  double ss_tr = 0;            /* sum sq of the treatments */
  double ss_smp = 0;           /* sum sq of the samples */
  double ss_err = 0;           /* sum sq of the error */
  double treatave[MAXTREAT];   /* means of the treatments */
  double sampleave[MAXSAMPLE]; /* means of the blocks */


  /* calculate mean of data */
  for (i=0, sum=0.0; i<numtreat; i++)
      for (j=0; j<numsample; j++)
          sum += data[i][j];
  mean = sum / (numtreat * numsample);

  /* now calculate the devsq of the data */
  for (i=0; i<numtreat; i++)
      for (j=0; j<numsample; j++)
          ss_tot += pow((data[i][j] - mean), 2);

  /* now calculate the means of the treatments */
  memset(treatave, 0, sizeof(treatave));
  for (i=0; i<numtreat; i++)
    {
      for (j=0; j<numsample; j++)
          treatave[i] += data[i][j];
      treatave[i] /= numsample;
    }

  /* calculate the sum square of the treatments */
  for (i=0; i<numtreat; i++)
      ss_tr += numsample * pow((treatave[i] - mean), 2);

  /* calculate the means of the samples */
  memset(sampleave, 0, sizeof(sampleave));
  for (i=0; i<numsample; i++)
    {
      for (j=0; j<numtreat; j++)
          sampleave[i] += data[j][i];
      sampleave[i] /= numtreat;
    }

  /* and calculate the sum square of the samples */
  for (i=0; i<numsample; i++)
      ss_smp += numtreat * pow((sampleave[i] - mean), 2);

  ss_err = ss_tot - ss_tr - ss_smp;
  mse = ss_err / ((numsample-1)*(numtreat-1));
  
  return mse;
}

/**********************************************************************/
/* calculates t values for vector of means passed in.  Assumes all
 * treatments have the same number of samples.
 */
void tvalues(double data[MAXTREAT][MAXSAMPLE], double means[MAXTREAT],
                              double tvals[COMPAR], double s)
{
  int i;               /* index variable */
  int j;               /* index variable */
  double mse = 0.0;    /* mean square of error */
  double me = 0.0;     /* mean of error */
  double mult[COMPAR]; /* contrast multiplier */

  if (unblocked == FALSE)
    {
      mse = meansquareerror(data);
      me = sqrt(mse);
    }
  for (i=0; i<numcompar; i++)
    {
      tvals[i] = 0.0;
      mult[i] = 0.0;
    }

  for (i=0; i<numcompar; i++)
      for (j=0; j<numtreat; j++)
          mult[i] += contrasts[i][j] * contrasts[i][j];

  for (i=0; i<numcompar; i++)
    {
      for (j=0; j<numtreat; j++)
        {
          if (unblocked == TRUE)
            tvals[i] += means[j]*contrasts[i][j] /
                                 (s * sqrt(mult[i]/numsample));
          else
            tvals[i] += means[j]*contrasts[i][j] /
                                 (me * sqrt(mult[i]/numsample));
        }
      tvals[i] = fabs(tvals[i]);
    }

  return;
}

/**********************************************************************/
/* Finds the max of the absolute values of the t-values passed in.
 * Used for single-step p-value adjustment
 */

double findmax(double tval[COMPAR])
{
  int i;
  double value = 0.0;
  double max = 0.0;

  for (i=0; i<numcompar; i++)
    {
      value = tval[i];
      if (value > max - DELTA)
          max = value;
    }
  return max;
}

/**********************************************************************/
/* calculates p values for vector of tvalues passed in.  If use_extpval
 * is TRUE, the p-value is passed to this function.  If use_extpval is
 * FALSE, the p-value is calculated from the t-value.  The means are
 * used only to sort the order of presentation.
 */
void pvalues(double tvals[COMPAR],
             double pvals[COMPAR], double means[MAXTREAT], int use_extpval)
{
  int i;                    /* index */
  int j;                    /* index */
  int comp;                 /* index */
  int df;                   /* degrees of freedom */
  doubsamples_t smeans;     /* sorted means */
  doubsamples_t *smeans_p;  /* pointer to sorted means */
  double pmatrix[MAXTREAT][MAXTREAT];

  if (use_extpval == FALSE)
    {
      if (unblocked == TRUE)
          df = (numsample - 1) * numtreat;
      else
          df = (numsample - 1) * (numtreat - 1);
      for (i=0; i<numcompar; i++)
          pvals[i] = 1.0 - student(df, tvals[i]);
    }

  /*** file pmatrix ***/
  comp = 0;
  for (i=0; i<numtreat-1; i++)
      for (j=i+1; j<numtreat; j++)
        {
          pmatrix[i][j] = pvals[comp];
          comp++;
        }
  for (i=0; i<numtreat; i++)
      for (j=0; j<i+1; j++)
        {
          if (i == j)
              pmatrix[i][j] = 1.0;
          pmatrix[i][j] = pmatrix[j][i];
        }

  /*** sort means ***/
  for (i=0; i<numtreat; i++)
    {
      smeans.data[i] = means[i];
      smeans.num[i] = i;
    }
  smeans_p = &smeans;
  doubsort(smeans_p, numtreat, MAX2MIN);

  /*** print out matrix ***/
  fprintf(stdout, "%-9s", "");
  for (i=1; i<numtreat; i++)
      fprintf(stdout, "%-9s", labels[smeans.num[i]]);
  fprintf(stdout, "\n");

  for (i=0; i<numtreat-1; i++)
    {
      fprintf(stdout, "%-9s", labels[smeans.num[i]]);
      for (j=1; j<numtreat; j++)
        {
          if (j <= i)
            {
              fprintf(stdout, "%s", "  -      ");
            }
          else
            {
              if (pmatrix[smeans.num[i]][smeans.num[j]] <= CRITSIG + DELTA)
                  fprintf(stdout, "%-5.3f*   ",
                          pmatrix[smeans.num[i]][smeans.num[j]]);
              else
                  fprintf(stdout, "%-5.3f    ",
                          pmatrix[smeans.num[i]][smeans.num[j]]);
            }
        }
      fprintf(stdout, "\n");
    }
  fprintf(stdout, "\n");
}

/*********************************************************************/
/* sorts (double) data in an array from min to max; direction is either
 * MIN2MAX or MAX2MIN
 */

void doubsort(doubsamples_t *sortedp, int n, int direction)
{
  int i;      /* index */
  int j;      /* index */
  struct
    {
      double data;
      int num;
    } temp;

  for (i=0; i<n-1; i++)
    {
      for (j=i+1; j<n; j++)
        {
          if (direction == MIN2MAX)
            {
              if (sortedp->data[j] < sortedp->data[i] + DELTA)
                {
                  temp.data = sortedp->data[i];
                  temp.num = sortedp->num[i];
                  sortedp->data[i] = sortedp->data[j];
                  sortedp->num[i] = sortedp->num[j];
                  sortedp->data[j] = temp.data;
                  sortedp->num[j] = temp.num;
                }
            }
          else  /* direction == MAX2MIN */
            {
              if (sortedp->data[j] > sortedp->data[i] - DELTA)
                {
                  temp.data = sortedp->data[i];
                  temp.num = sortedp->num[i];
                  sortedp->data[i] = sortedp->data[j];
                  sortedp->num[i] = sortedp->num[j];
                  sortedp->data[j] = temp.data;
                  sortedp->num[j] = temp.num;
                }
            }
        }
    }
}

/**********************************************************************/
/* Reads the input file.  Format is as described at the top of this
 * file.  Lines starting with percent (%) or which are blank are
 * skipped over.  Tokens are separated by commas, spaces, tabs, line
 * returns, and newlines.  Treatment labels are truncated to MAXLABEL
 * characters.  If number of listeners exceed MAXSAMPLE or number of
 * treatments exceed MAXTREAT, the program exits with an error message.
 */

void read_input(void)
{
  char line[MAXCHAR];          /* string buffer */
  char *p;                     /* generic pointer */
  char buf[MAXCHAR];           /* string buffer */
  int labels_parsed = FALSE;   /* flag for labels parsed */
  int count;                   /* count of data */
  int linenum = 0;             /* line number */

  ifp = fopen(infile, "r");
  if (ifp == NULL) 
    {
      fprintf(stderr, "Can't open input file %s\n", infile);
      exit(1);
    }
  else
    {
      printf("Input file : %s\n", infile);
    }

  numtreat = 0;
  numsample = 0;

  while (fgets(line, MAXCHAR, ifp))
    {
      p = strtok(line, ", \t\r\n");
      linenum++;

      /* ignore comments and blank lines */
      if (p)
        {
          strncpy(buf, p, 1);
          buf[1] = '\0';
          if (buf[0] == '%')
          continue;
        }
      else
          continue;

      if (labels_parsed == FALSE)
        {
          strcpy(labels[0], p);
          labels[0][MAXLABEL] = '\0';
          numtreat++;

          while ( (p = strtok(NULL, ", \t\r\n")) != NULL)
            {
              strcpy(labels[numtreat], p);
              labels[numtreat][MAXLABEL] = '\0';
              numtreat++;
            }
          if (numtreat > MAXSAMPLE)
            {
              fprintf(stderr, "Max treatements allowed: %d\n", MAXTREAT);
              exit(1);
            }
          labels_parsed = TRUE;
          continue;
        }

      /* parse data */
      count = 0;
      strcpy(buf, p);
      or_data[count][numsample] = atof(buf);
      count++;

      while ( (p = strtok(NULL, ", \t\r\n")) != NULL)
        {
          strcpy(buf, p);
          or_data[count][numsample] = atof(buf);
          count++;
        }
      numsample++;
      if (numsample > MAXSAMPLE)
        {
          fprintf(stdout, "Max number of listeners allowed:  %d\n", MAXSAMPLE);
          exit(1);
        }
    }

  numcompar = numtreat * (numtreat - 1) / 2;
  printf("Read %d treatments, %d samples\n\n", numtreat, numsample);

  return;
}

/*********************************************************************/
/* Resamples the data.
 *
 * If bootstrap resampling is chosen, the function will generate a
 * resampled matrix of data, first by randomly choosing samples
 * (listeners) with replacement, then for each listener, randomly
 * choosing treatment values (codec ratings) from that listener's
 * pool of ratings, also with replacement.
 *
 * If permutation resampling is chosen, the function will generate
 * a resampled matrix of data by shuffling the treatment values for
 * each listener.  Thus, no listeners are replaced, and no ratings
 * for each listener are replaced either.
 *
 * if unblocked is TRUE, function will perform resampling by pooling
 * all values over the entire listener x codec space.
 */

void resample_data(double ordata[MAXTREAT][MAXSAMPLE],
                   double rsdata[MAXTREAT][MAXSAMPLE])
{
  int i, j, k, l;                   /* index variables */
  int picksample = 0;               /* index of sample picked */
  int picktreat;                    /* index of treatment picked */
  int picked[MAXTREAT][MAXSAMPLE];  /* TRUE if data has been picked */
  int picks[MAXTREAT];              /* pool of treatments to be picked from */

  for (i = 0; i < numsample; i++)
    {
      switch (resamplemode)
        {
          case PERM:
              if (unblocked == FALSE)
                {
                  picksample = i;
                  memset(picked, 0, sizeof(picked));
                  for (j = 0; j < numtreat; j++)
                    {
                      if (j == 0) 
                          picktreat = (int)((numtreat)*(randomMT()/(ULONG_MAX+1.0)));
                      else
                        {
                          /* gather picks */
                          for (k = 0, l = 0; k < numtreat; k++)
                            {
                              if (!picked[k][i])
                                {
                                  picks[l] = k;
                                  l++;
                                }
                            }

                          if (l == 1)
                              picktreat = picks[0];
                          else
                              picktreat = picks[(int)((l)*(randomMT()/(ULONG_MAX+1.0)))];
                        }
                      picked[picktreat][i] = TRUE;
                      rsdata[j][i] = ordata[picktreat][picksample];
                    }  /* for j */
                }
              else  /* unblocked == TRUE */
                {
                  /* NOT IMPLEMENTED YET */
                }
              break;
          case BOOT:
              if (unblocked == FALSE)
                  picksample = (int)((numsample)*(randomMT()/(ULONG_MAX+1.0)));
              for (j = 0; j < numtreat; j++)
                {
                  if (unblocked == TRUE)
                      picksample = (int)((numsample)*(randomMT()/(ULONG_MAX+1.0)));
                  picktreat = (int)((numtreat)*(randomMT()/(ULONG_MAX+1.0)));
                  rsdata[j][i] = ordata[picktreat][picksample];
                }
              break;
          default:
              break;
        }
    }  /* for i */
}

/**********************************************************************/

void usage_error(void)
{
  fprintf(stdout,
  "Usage: bootstrap [options] <infile>\n");
  fprintf(stdout,
  "  -h               Usage\n");
  fprintf(stdout,
  "  -n               Number of resamples to derive adjusted p-values\n"
  "                     default: 10,000\n");
  fprintf(stdout,
  "  -perm            Perform permutation resampling\n"
  "                     default: bootstrap\n");
  fprintf(stdout,
  "  -adj <mode>      p-value adjustment mode\n"
  "                     noadj:   make no p-value adjustments\n"
  "                     single:  single-step\n"
  "                     default: free step-down\n");
  fprintf(stdout,
  "  -unbl            Perform unblocked resampling\n"
  "                     default: blocked\n");
  exit(1);
}

/**********************************************************************/
/* process command line options
 * Returns INCR if switch option is followed by another argument;
 * Otherwise returns NO_INCR
 */

int process_options(int argnum, char *argv[])
{
  char option[MAXCHAR];       /* option */
  char optionvalue[MAXCHAR];  /* value of option */

  sprintf(option, "%s", *(argv + argnum));

  if (strcmp(option, "-h") == 0)
  usage_error();
  else
    {
      if (strcmp(option, "-n") == 0)
        {
          sprintf(optionvalue, "%s", *(argv + argnum + 1));
          npadj = atol(optionvalue);
          if (npadj <= 0)
            {
              fprintf(stdout, "Must specify an n greater than 0\n");
              exit(1);
            }
          return INCR;
        }
      if (strcmp(option, "-perm") == 0)
        {
          sprintf(optionvalue, "%s", *(argv + argnum + 1));
          resamplemode = PERM;
          return NO_INCR;
        }
      if (strcmp(option, "-adj") == 0)
        {
          sprintf(optionvalue, "%s", *(argv + argnum + 1));
          if (strcmp(optionvalue, "noadj") == 0)
            {
              adjustmode = NOADJ;
              return INCR;
            }
          if (strcmp(optionvalue, "single") == 0)
            {
              adjustmode = SINGLE;
              return INCR;
            }
          else
            {
              fprintf(stdout, "%s is not a valid parameter for -adj\n", optionvalue);
              exit(1);
            }
        }
      if (strcmp(option, "-unbl") == 0)
        {
          sprintf(optionvalue, "%s", *(argv + argnum + 1));
          unblocked = TRUE;
          return NO_INCR;
        }

      /* if you've come this far without finding an option,
       * it's an error
       */
      fprintf(stdout, "Option '%s' is not recognized\n", option);
      usage_error();
    }

  return NO_INCR;
}

/**********************************************************************/
/* user must enter input filename
 */

int parse_command_line(int argc, char *argv[])
{
  int i;
  int increment = FALSE;   /* increment argument counter */

  if (argc < 2)
    {
      fprintf(stdout, "Not enough arguments specified\n");
      usage_error();
    }

  for (i=1; i<argc; i++)
    {
      if (argv[i][0] == '-')
        {
          increment = process_options(i, argv);
          if (increment)
            {
              i++;
              continue;
            }
        }
      else
          if (infile == NULL)
              infile = argv[i];
          else
            {
              fprintf(stdout, "Too many arguments specified\n");
              usage_error();
            }
    }

  if (infile == NULL)
    {
      fprintf(stdout, "Input filename not specified\n");
      usage_error();
    }

  return 0;
}

/**********************************************************************/
/* prints version to screen or to file
 */

void version(FILE *output)
{
  fprintf(output, "BOOTSTRAP version %s,  "
                  "Copyright (c) by Garf and ff123\n", VERSION);
  fprintf(output, "http://sjeng.sourceforge.net/, http://ff123.net/\n");
  fprintf(output, "http://www.gnu.org/copyleft/lesser.html\n\n");
}

/*********************************************************************/

int main(int argc, char *argv[])
{
  long i;
  int j;
  int k;
  double maxt_free = 0.0;     /* max t-value for free step-down */
  double maxt_single = 0.0;   /* max t-value for single step */

  version(stdout);
  init();
  parse_command_line(argc, argv);

  /* unblocked permutation resampling not yet implemented */
  if ((unblocked == TRUE) && (resamplemode == PERM))
    {
      fprintf(stdout, "Unblocked permutation resampling not implemented.\n");
      exit(1);
    }

  /* free step down for permutation resampling is not allowed */
  if ((resamplemode == PERM) && (adjustmode == FREE))
    {
      fprintf(stdout, "Free step-down is not allowed with permutation resampling\n");
      fprintf(stdout, "Use -adj single\n");
      exit(1);
    }

  seedMT(SEED);
  read_input();
  create_default_contrasts();

#if 0
  add_test_contrasts();
#endif

  treatmeans(or_data, or_means);
  raw_s = residualmean(or_data, or_means);
  tvalues(or_data, or_means, rawtval, raw_s);
  printf("                            Unadjusted p-values\n");
  pvalues(rawtval, rawpval, or_means, CALC_PVAL);

  printf("Each '.' is 1,000 resamples.  Each '+' is 10,000 resamples\n");

  for (i=1; i<npadj+1; i++)
    {
      if ((i % 10000) == 0)
          printf("+");
      else
          if ((i % 1000) == 0)
              printf(".");
      resample_data(or_data, rs_data);
      treatmeans(rs_data, rs_means);
      res_s = residualmean(rs_data, rs_means);
      tvalues(rs_data, rs_means, restval, res_s);

      for (k=0; k<numcompar; k++)
        {
          min2max_tval.data[k] = rawtval[k];
          max2min_tval.data[k] = rawtval[k];
          min2max_tval.num[k] = k;
          max2min_tval.num[k] = k;
        }
      min2max_tvalp = &min2max_tval;
      max2min_tvalp = &max2min_tval;
      doubsort(min2max_tvalp, numcompar, MIN2MAX);
      doubsort(max2min_tvalp, numcompar, MAX2MIN);

      for (j=0, maxt_free=0; j<numcompar; j++)
        {
          switch (adjustmode)
            {
              case NOADJ:
                  if (restval[j] >= rawtval[j] - DELTA)
                      hits[j]++;
                  break;
              case SINGLE:
                  maxt_single = findmax(restval);
                  if (maxt_single >= rawtval[j] - DELTA)
                      hits[j]++;
                  break;
              case FREE:
                  maxt_free = MAX2VAL(restval[min2max_tval.num[j]], maxt_free);
                  if (maxt_free >= rawtval[min2max_tval.num[j]] - DELTA)
                      hits[min2max_tval.num[j]]++;
                  break;
              default:
                  break;
            }
        }
    }
  printf("\n\n");

  for (j=0; j<numcompar; j++)
    {
      respval[j] = (double)hits[j] / npadj;
      if (adjustmode == FREE)
        {
          /* enforce monoticity */
          if (j>0)
              if (respval[max2min_tval.num[j]] < respval[max2min_tval.num[j-1]] + DELTA)
                  respval[max2min_tval.num[j]] = respval[max2min_tval.num[j-1]];
        }
    }

  printf("                             Adjusted p-values\n");
  pvalues(NULL, respval, or_means, PASS_PVAL);

  if (ifp != NULL) fclose(ifp);
  return 0;
}
