/* bootstrap.h
 *
 * written and copyright by Garf and ff123
 * http://sjeng.sourceforge.net/
 * http://ff123.net/
 *
 * Project header file
 *
 *-------------------- GNU Lesser General Public License -------------------
 * http://www.gnu.org/copyleft/lesser.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *********************************************************************/

#ifndef BOOTSTRAP_H
  #define BOOTSTRAP_H

  #include <stdio.h>
  #include <stdlib.h>
  #include <string.h>
  #include <limits.h>
  #include <math.h>

  #define VERSION   "0.5, Nov 18, 2001"

  #define SEED             4321   /* seed for random number gen */
  #define MAXTREAT         8      /* max number of treatments (codecs) */
  #define EXTRA_CONTRASTS  6      /* extra tests as desired by user */
  #define MAXSAMPLE        50     /* max number of samples (listeners) */
  #define MAXCHAR          255    /* max number of characters in a string */
  #define MAXLABEL         8      /* max number of characters in a treatment label */
  #define COMPAR ((MAXTREAT*(MAXTREAT-1))/2 + EXTRA_CONTRASTS)
                                  /* number of comparisons */
  #define MAX2VAL(A,B) ( (A) > (B) ? (A):(B))
  #define CRITSIG   0.05          /* critical significance */

  #define DELTA 1E-15             /* used when making floating-point comparisons */
  enum {FALSE, TRUE};
  enum                            /* resampling mode */
    {
      PERM,                       /* permutation resampling */
      BOOT,                       /* bootstrap resampling */
    };
  enum                            /* resampling mode */
    {
      NOADJ,                      /* no p-value adjustment */
      SINGLE,                     /* single-step p-value adjustment */
      FREE,                       /* free step-down p-value adjustment */
    };
  enum {MIN2MAX, MAX2MIN};        /* direction of sort algorithm */
  enum {NO_INCR, INCR};           /* used for parsing command line */
  enum
    {
      CALC_PVAL,                  /* calculate p-value using tvals */
      PASS_PVAL,                  /* pass in p-values */
    };

  FILE *ifp;                      /* input file pointer */
  int resamplemode;               /* resampling mode */
  int adjustmode;                 /* p-value adjustment mode */
  int numtreat;                   /* number of treatments (codecs) */
  int numsample;                  /* number of samples (listeners) */
  int numcompar;                  /* number of t comparisons */
  long npadj;                     /* number of resamples to derive adjusted p */
  int unblocked;                  /* if TRUE, does not block per listener */
  double raw_s;                   /* raw residual mean */
  double res_s;                   /* resampled residual mean */

  double or_data[MAXTREAT][MAXSAMPLE];  /* raw data */
  double rs_data[MAXTREAT][MAXSAMPLE];  /* resampled data */
  char labels[MAXTREAT][MAXCHAR];       /* names of the treatments */
  double or_means[MAXTREAT];            /* means of raw data */
  double rs_means[MAXTREAT];            /* resampled means */
  double rawtval[COMPAR];               /* raw t values */
  double restval[COMPAR];               /* resampled t values */
  double rawpval[COMPAR];               /* raw p values */
  double respval[COMPAR];               /* resampled p values */
  long hits[COMPAR];                    /* t-value hits */
  int contrasts[COMPAR][MAXTREAT];      /* contrasts matrix */

  char *infile;                   /* input filename */

  typedef struct
    {
      long data[COMPAR];          /* (long) data to be sorted */
      int num[COMPAR];            /* numbering of data */
    } longsamples_t;

  typedef struct
    {
      double data[COMPAR];        /* (double) data to be sorted */
      int num[COMPAR];            /* numbering of data */
    } doubsamples_t;

  typedef unsigned long uint32;   /* for use with random functions */

  doubsamples_t min2max_tval;     /* sorted t-values */
  doubsamples_t *min2max_tvalp;   /* pointer to sorted t-values */
  doubsamples_t max2min_tval;     /* sorted t-values */
  doubsamples_t *max2min_tvalp;   /* pointer to sorted t-values */

/* -------------------- function prototypes ------------------- */

  /* bootstrap.c */
  void init(void);
  void treatmeans(double data[MAXTREAT][MAXSAMPLE],
                    double treatmeans[MAXTREAT]);
  double residualmean(double data[MAXTREAT][MAXSAMPLE],
                    double means[MAXTREAT]);
  double meansquareerror(double data[MAXTREAT][MAXSAMPLE]);
  void tvalues(double data[MAXTREAT][MAXSAMPLE], double means[MAXTREAT],
                              double tvals[COMPAR], double s);
  double findmax(double tval[COMPAR]);
  void pvalues(double tvals[COMPAR],
       double pvals[COMPAR], double means[MAXTREAT], int use_extpval);
  void doubsort(doubsamples_t *sortedp, int n, int direction);
  void print_rank_sums(double rks[MAXTREAT]);
  void read_input(void);
  void resample_data(double ordata[MAXTREAT][MAXSAMPLE],
                   double rsdata[MAXTREAT][MAXSAMPLE]);
  void usage_error(void);
  int process_options(int argnum, char *argv[]);
  int parse_command_line(int argc, char *argv[]);
  void version(FILE *output);

  /* contrasts.c */
  void create_default_contrasts(void);
  void add_test_contrasts(void);

  /* fisher.c */
  double student(int df, double t);
  double critt(int df, double p);

  /* random.c */
  void seedMT(uint32 seed);
  uint32 reloadMT(void);
  uint32 randomMT(void);

#endif  /* BOOTSTRAP_H */